"""Etekcity/Valceno Smart Light Bulbs.

This module provides classes for the following Etekcity/Valceno smart lights:

    1. ESL100: Dimmable Bulb
    2. ESL100CW: Tunable White Bulb
    3. XYD0001: RGB Bulb
    4. ESL100MC: Multi-Color Bulb

Attributes:
    feature_dict (dict): Dictionary of bulb models and their supported features.
        Defines the class to use for each bulb model and the list of features
    bulb_modules (dict): Dictionary of bulb models as keys and their associated classes
        as string values.

Note:
    The bulb module is built from the `feature_dict` dictionary and used by the
    `vesync.object_factory` and tests to determine the class to instantiate for
    each bulb model. These classes should not be instantiated manually.

Examples:
    The following example shows the structure of the `feature_dict` dictionary:
    ```python

    feature_dict = {
        'ESL100MC': { # device_type attribute
            'module': 'VeSyncBulbESL100MC', # String name of the class to instantiate
            'features': ['dimmable', 'rgb_shift'], # List of supported features
            'color_model': 'rgb' # Color model used by the bulb (rgb, hsv, none)
        }
    }
    ```

"""
from __future__ import annotations
import logging
import json
from typing import Union, Optional, NamedTuple, TYPE_CHECKING
from abc import ABCMeta, abstractmethod
from pyvesync.helpers import Helpers as helpers, Color
from pyvesync.vesyncbasedevice import VeSyncBaseDevice


if TYPE_CHECKING:
    from pyvesync import VeSync

logger = logging.getLogger(__name__)

NUMERIC_T = Optional[Union[int, float, str]]

# --8<-- [start:feature_dict]

feature_dict: dict = {
    'ESL100':
        {
            'module': 'VeSyncBulbESL100',
            'features': ['dimmable'],
            'color_model': 'none'
        },
    'ESL100CW':
        {
            'module': 'VeSyncBulbESL100CW',
            'features': ['dimmable', 'color_temp'],
            'color_model': 'none'
        },
    'XYD0001':
        {
            'module': 'VeSyncBulbValcenoA19MC',
            'features': ['dimmable', 'color_temp', 'rgb_shift'],
            'color_model': 'hsv'
        },
    'ESL100MC':
        {
            'module': 'VeSyncBulbESL100MC',
            'features': ['dimmable', 'rgb_shift'],
            'color_model': 'rgb'
        }
}

# --8<-- [end:feature_dict]

bulb_modules: dict = {k: v['module'] for k, v in feature_dict.items()}

__all__: list = list(bulb_modules.values()) + ["bulb_modules", "VeSyncBulb"]


def pct_to_kelvin(pct: float, max_k: int = 6500, min_k: int = 2700) -> float:
    """Convert percent to kelvin."""
    return ((max_k - min_k) * pct / 100) + min_k


class VeSyncBulb(VeSyncBaseDevice):
    """Base class for VeSync Bulbs.

    Abstract base class to provide methods for controlling and
    getting details of VeSync bulbs. Inherits from
    [`VeSyncBaseDevice`][pyvesync.vesyncbasedevice.VeSyncBaseDevice]. This class
    should not be used directly for devices, but rather subclassed for each.

    Attributes:
        brightness (int): Brightness of bulb (0-100).
        color_temp_kelvin (int): White color temperature of bulb in Kelvin.
        color_temp_pct (int): White color temperature of bulb in percent (0-100).
        color_hue (float): Color hue of bulb (0-360).
        color_saturation (float): Color saturation of bulb in percent (0-100).
        color_value (float): Color value of bulb in percent (0-100).
        color (Color): Color of bulb in the form of a dataclass with two namedtuple
            attributes - `hsv` & `rgb`. See [pyvesync.helpers.Color][].
    """

    __metaclass__ = ABCMeta

    def __init__(self, details: dict[str, str | list],
                 manager: VeSync) -> None:
        """Initialize VeSync smart bulb base class."""
        super().__init__(details, manager)
        self._brightness = 0
        self._color_temp = 0
        self._color_value = float(0)
        self._color_hue = float(0)
        self._color_saturation = float(0)
        self._color_mode: str = ''   # possible: white, color, hsv
        self._color: Color | None = None
        self.features: list | None = feature_dict.get(
            self.device_type, {}).get('features')
        if self.features is None:
            logger.error("No configuration set for - %s", self.device_name)
            raise KeyError
        self._rgb_values = {
            'red': 0,
            'green': 0,
            'blue': 0
        }

    @property
    def brightness(self) -> int:
        """Return brightness of vesync bulb.

        Returns:
            int: Brightness of bulb (0-100).
        """
        if self.dimmable_feature and self._brightness is not None:
            return self._brightness
        return 0

    @property
    def color_temp_kelvin(self) -> int:
        """Return white color temperature of bulb in Kelvin.

        Converts the color temperature in percent to Kelvin using
        the `pct_to_kelvin` function.

        Returns:
            int: White color temperature of bulb in Kelvin (2700 - 6500).

        Notes:
            This returns 0 for bulbs that do not have color temperature support.
        """
        if self.color_temp_feature and self._color_temp is not None:
            return int(pct_to_kelvin(self._color_temp))
        return 0

    @property
    def color_temp_pct(self) -> int:
        """Return white color temperature of bulb in percent (0-100).

        Subclasses that use this method, should calculate the color temeprature
        in percent regardless of how the API returns the value.
        """
        if self.color_temp_feature and self._color_temp is not None:
            return int(self._color_temp)
        return 0

    @property
    def color_hue(self) -> float:
        """Return color hue (HSV colorspace) of bulb.

        Returns hue from the `color` attribute. (0-360)

        Returns:
            float: Color hue of bulb in HSV colorspace.

        Notes:
            This returns 0 for bulbs that do not have color support.
        """
        if self.rgb_shift_feature and self._color is not None:
            return self._color.hsv.hue
        return 0

    @property
    def color_saturation(self) -> float:
        """Return color saturation (HSV colorspace) of bulb in percent.

        Return saturation from the `color` attribute (0-100).

        Returns:
            float: Color saturation of bulb in percent (0-100).

        Notes:
            This returns 0 for bulbs that do not have color
        """
        if self.rgb_shift_feature and self._color is not None:
            return self._color.hsv.saturation
        return 0

    @property
    def color_value(self) -> float:
        """Return color value (HSV colorspace) of bulb in percent.

        Returns the color from from the `color` attribute (0-100).

        Returns:
            float: Color value of bulb in percent (0-100).

        Notes:
            This returns 0 for bulbs that do not have color support.
        """
        if self.rgb_shift_feature and self._color is not None:
            return self._color.hsv.value
        return 0

    @property
    # pylint: disable-next=differing-param-doc # DOCUMENTATION FOR SETTER
    def color(self) -> Color | None:
        # pylint: disable=differing-type-doc
        """Set color property based on rgb or hsv values.

        Pass either red, green, blue or hue, saturation, value.

        Args:
            red (float): Red value of RGB color, 0-255
            green (float): Green value of RGB color, 0-255
            blue (float): Blue value of RGB color, 0-255
            hue (float): Hue value of HSV color, 0-360
            saturation (float): Saturation value of HSV color, 0-100
            value (float): Value (brightness) value of HSV color 0-100

        Returns:
            Color: Color dataclass with hsv and rgb named tuple attributes.
        """
        if self.rgb_shift_feature is True and self._color is not None:
            return self._color
        return None

    @color.setter
    def color(self, red: float | None = None,
              green: float | None = None,
              blue: float | None = None,
              hue: float | None = None,
              saturation: float | None = None,
              value: float | None = None) -> None:
        """Set color property based on rgb or hsv values."""
        self._color = Color(red=red, green=green, blue=blue,
                            hue=hue, saturation=saturation, value=value)

    @property
    def color_hsv(self) -> NamedTuple | None:
        """Return color of bulb as [hsv named tuple][pyvesync.helpers.HSV].

        Notes:
            Returns `None` for bulbs that do not have color support.
        """
        if self.rgb_shift_feature is True and self._color is not None:
            return self._color.hsv
        return None

    @property
    def color_rgb(self) -> NamedTuple | None:
        """Return color of bulb as [rgb named tuple][pyvesync.helpers.RGB].

        Notes:
            Returns `None` for bulbs that do not have color support.
        """
        if self.rgb_shift_feature is True and self._color is not None:
            return self._color.rgb
        return None

    @property
    def color_mode(self) -> str | None:
        """Return color mode of bulb. Possible values are none, hsv or rgb.

        Notes:
            This is a read-only property. Color mode is defined in
            the [`feature_dict`][pyvesync.vesyncbulb.feature_dict].
        """
        if self.rgb_shift_feature and self._color_mode is not None:
            return str(self._color_mode)
        return None

    @property
    def dimmable_feature(self) -> bool:
        """Return true if dimmable bulb."""
        return (self.features is not None and 'dimmable' in self.features)

    @property
    def color_temp_feature(self) -> bool:
        """Checks if the device has the ability to change color temperature.

        Returns:
            bool: True if the device supports changing color temperature.
        """
        return (self.features is not None and 'color_temp' in self.features)

    @property
    def rgb_shift_feature(self) -> bool:
        """Checks if the device is multicolor.

        Returns:
            bool: True if the device supports changing color.
        """
        return (self.features is not None and 'rgb_shift' in self.features)

    def _validate_rgb(self, red: NUMERIC_T = None,
                      green: NUMERIC_T = None,
                      blue: NUMERIC_T = None) -> Color:
        """Validate RGB values."""
        rgb_dict = {'red': red, 'green': green, 'blue': blue}
        for clr, val in rgb_dict.items():
            if val is None:
                rgb_dict[clr] = getattr(self._rgb_values, clr)
            else:
                rgb_dict[clr] = val
        return Color(red=rgb_dict['red'],
                     green=rgb_dict['green'],
                     blue=rgb_dict['blue'])

    def _validate_hsv(self, hue: NUMERIC_T = None,
                      saturation: NUMERIC_T = None,
                      value: NUMERIC_T = None) -> Color:
        """Validate HSV Arguments."""
        hsv_dict = {'hue': hue, 'saturation': saturation, 'value': value}
        for clr, val in hsv_dict.items():
            if val is None and self._color is not None:
                hsv_dict[clr] = getattr(self._color.hsv, clr)

        if hue is not None:
            valid_hue = self._validate_any(hue, 1, 360, 360)
        elif self._color is not None:
            valid_hue = self._color.hsv.hue
        else:
            logger.debug("No current hue value, setting to 0")
            valid_hue = 360
        hsv_dict['hue'] = valid_hue
        for itm, val in {'saturation': saturation, 'value': value}.items():
            if val is not None:
                valid_item = self._validate_any(val, 1, 100, 100)
            elif self.color is not None:
                valid_item = getattr(self.color.hsv, itm)
            else:
                logger.debug("No current %s value, setting to 0", itm)
                valid_item = 100
            hsv_dict[itm] = valid_item
        return Color(hue=hsv_dict['hue'], saturation=hsv_dict['saturation'],
                     value=hsv_dict['value'])

    def _validate_brightness(self, brightness: float | str,
                             start: int = 0, stop: int = 100) -> int:
        """Validate brightness value."""
        try:
            brightness_update: int = max(start, (min(stop, int(
                                                round(float(brightness), 2)))))
        except (ValueError, TypeError):
            brightness_update = self.brightness if self.brightness is not None else 100
        return brightness_update

    def _validate_color_temp(self, temp: int, start: int = 0, stop: int = 100) -> int:
        """Validate color temperature."""
        try:
            temp_update = max(start, (min(stop, int(
                round(float(temp), 0)))))
        except (ValueError, TypeError):
            temp_update = self._color_temp if self._color_temp is not None else 100
        return temp_update

    @staticmethod
    def _validate_any(value: NUMERIC_T,
                      start: NUMERIC_T = 0,
                      stop: NUMERIC_T = 100,
                      default: float = 100) -> float:
        """Validate any value."""
        try:
            value_update = max(float(start), (min(float(stop), round(float(value), 2))))  # type: ignore[arg-type]  # noqa
        except (ValueError, TypeError):
            value_update = default
        return value_update

    @abstractmethod
    def set_status(self) -> bool:
        """Set vesync bulb attributes(brightness, color_temp, etc).

        This is a helper function that is called by the direct `set_*` methods,
        such as `set_brightness`, `set_rgb`, `set_hsv`, etc.

        Returns:
            bool : True if successful, False otherwise.
        """

    @abstractmethod
    def get_details(self) -> None:
        """Get vesync bulb details.

        This is a legacy function to update devices, **updates should be
        called by `update()`**

        Returns:
            None
        """

    @abstractmethod
    def _interpret_apicall_result(self, response: dict) -> None:
        """Update bulb status from any api call response."""

    @abstractmethod
    def toggle(self, status: str) -> bool:
        """Toggle mode of vesync lightbulb.

        Helper function called by `turn_on()` and `turn_off()`.

        Args:
            status (str): 'on' or 'off'

        Returns:
            bool: True if successful, False otherwise.
        """

    @abstractmethod
    def get_config(self) -> None:
        """Call api to get configuration details and firmware.

        Populates the `self.config` attribute with the response.

        Returns:
            None

        Note:
            The configuration attribute `self.config` is structured as follows:
            ```python
            {
                'current_firmware_version': '1.0.0',
                'latest_firmware_version': '1.0.0',
                'maxPower': '560',
                'threshold': '1000',
                'power_protection': 'on',
                'energy_saving_status': 'on'
            }
            ```
        """

    def set_hsv(self,
                hue: NUMERIC_T,
                saturation: NUMERIC_T,
                value: NUMERIC_T) -> bool | None:
        """Set HSV if supported by bulb.

        Args:
            hue (NUMERIC_T): Hue 0-360
            saturation (NUMERIC_T): Saturation 0-100
            value (NUMERIC_T): Value 0-100

        Returns:
            bool: True if successful, False otherwise.
        """
        if self.rgb_shift_feature is False:
            logger.debug("HSV not supported by bulb")
            return False
        return bool(hue and saturation and value)

    def set_rgb(self, red: NUMERIC_T = None,
                green: NUMERIC_T = None,
                blue: NUMERIC_T = None) -> bool:
        """Set RGB if supported by bulb.

        Args:
            red (NUMERIC_T): Red 0-255
            green (NUMERIC_T): green 0-255
            blue (NUMERIC_T): blue 0-255

        Returns:
            bool: True if successful, False otherwise.
        """
        if self.rgb_shift_feature is False:
            logger.debug("RGB not supported by bulb")
            return False
        return bool(red and green and blue)

    def turn_on(self) -> bool:
        """Turn on vesync bulbs.

        Calls `toggle('on')`.

        Returns:
            bool : True if successful, False otherwise.
        """
        if self.toggle('on'):
            self.device_status = 'on'
            return True
        logger.warning('Error turning %s on', self.device_name)
        return False

    def turn_off(self) -> bool:
        """Turn off vesync bulbs.

        Calls `toggle('off')`.

        Returns:
            bool : True if successful, False otherwise.
        """
        if self.toggle('off'):
            self.device_status = 'off'
            return True
        logger.warning('Error turning %s off', self.device_name)
        return False

    def update(self) -> None:
        """Update bulb details.

        Calls `get_details()` method to retrieve status from API and
        update the bulb attributes. `get_details()` is overriden by subclasses
        to hit the respective API endpoints.
        """
        self.get_details()

    def display(self) -> None:
        """Return formatted bulb info to stdout."""
        super().display()
        if self.connection_status == 'online':
            disp = []  # initiate list
            if self.dimmable_feature:
                disp.append(('Brightness: ', str(self.brightness), '%'))
            if self.color_temp_feature:
                disp.append(('White Temperature Pct: ',
                             str(self.color_temp_pct), '%'))
                disp.append(('White Temperature Kelvin: ',
                             str(self.color_temp_kelvin), 'K'))
            if self.rgb_shift_feature and self.color is not None:
                disp.append(('ColorHSV: ', helpers.named_tuple_to_str(
                    self.color.hsv), ''))
                disp.append(('ColorRGB: ', helpers.named_tuple_to_str(
                    self.color.rgb), ''))
                disp.append(('ColorMode: ', str(self.color_mode), ''))
            if len(disp) > 0:
                for line in disp:
                    print(f'{line[0]:.<30} {line[1]} {line[2]}')

    def displayJSON(self) -> str:
        """Return bulb device info in JSON format.

        Returns:
            str: JSON formatted string of bulb details.

        Example:
            ```json
            {
                "deviceName": "Bulb",
                "deviceStatus": "on",
                "connectionStatus": "online",
                "Brightness": "100%",
                "WhiteTemperaturePct": "100%",
                "WhiteTemperatureKelvin": "6500K",
                "ColorHSV": "{"hue": 0, "saturation": 0, "value": 0}",
                "ColorRGB": "{"red": 0, "green": 0, "blue": 0}",
                "ColorMode": "hsv"
            }
            ```
        """
        sup = super().displayJSON()
        sup_val = json.loads(sup)
        if self.connection_status == 'online':
            if self.dimmable_feature:
                sup_val.update({'Brightness': str(self.brightness)})
            if self.color_temp_feature:
                sup_val.update(
                    {'WhiteTemperaturePct': str(self.color_temp_pct)})
                sup_val.update(
                    {'WhiteTemperatureKelvin': str(self.color_temp_kelvin)})
            if self.rgb_shift_feature:
                if self.color_hsv is not None:
                    sup_val.update({'ColorHSV': json.dumps(
                        self.color_hsv._asdict())})
                if self.color_rgb is not None:
                    sup_val.update({'ColorRGB': json.dumps(
                        self.color_rgb._asdict())})
                sup_val.update({'ColorMode': str(self.color_mode)})
        return json.dumps(sup_val, indent=4)

    @property
    def color_value_rgb(self) -> NamedTuple | None:
        """Legacy Method .... Depreciated."""
        if self._color is not None:
            return self._color.rgb
        return None

    @property
    def color_value_hsv(self) -> NamedTuple | None:
        """Legacy Method .... Depreciated."""
        if self._color is not None:
            return self._color.hsv
        return None


class VeSyncBulbESL100MC(VeSyncBulb):
    """Etekcity ESL100 Multi Color Bulb device instance.

    Inherits from [VeSyncBulb][pyvesync.vesyncbulb.VeSyncBulb]
    and [VeSyncBaseDevice][pyvesync.vesyncbasedevice.VeSyncBaseDevice].

    Attributes:
        device_status (str): Status of bulb, either 'on' or 'off'.
        connection_status (str): Connection status of bulb, either 'online' or 'offline'.
        details (dict): Dictionary of bulb state details.
        brightness (int): Brightness of bulb (0-100).
        color_temp_kelvin (int): White color temperature of bulb in Kelvin.
        color_temp_pct (int): White color temperature of bulb in percent (0-100).
        color_hue (float): Color hue of bulb (0-360).
        color_saturation (float): Color saturation of bulb in percent (0-100).
        color_value (float): Color value of bulb in percent (0-100).
        color (Color): Color of bulb in the form of a dataclass with
            two named tuple attributes - `hsv` & `rgb`. See [pyvesync.helpers.Color][].

    Notes:
        The details dictionary contains the device information retreived by the
        `update()` method:
        ```python
        details = {
            'brightness': 50,
            'colorMode': 'rgb',
            'color' : Color(red=0, green=0, blue=0)
        }
        ```
        See pyvesync.helpers.Color for more information on the Color dataclass.
    """

    def __init__(self, details: dict[str, str | list], manager: VeSync) -> None:
        """Instantiate ESL100MC Multicolor Bulb.

        Args:
            details (dict): Dictionary of bulb state details.
            manager (VeSync): Manager class used to make API calls.
        """
        super().__init__(details, manager)
        self.details: dict = {}

    def get_details(self) -> None:
        head = helpers.bypass_header()
        body = helpers.bypass_body_v2(self.manager)
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'getLightStatus',
            'source': 'APP',
            'data': {}
        }

        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            method='post',
            headers=head,
            json_object=body,
        )
        if not isinstance(r, dict) or not isinstance(r.get('result'), dict) \
                or r.get('code') != 0:
            logger.debug("Error in bulb response")
            return
        outer_result = r.get('result', {})
        inner_result = outer_result.get('result')

        if inner_result is None or outer_result.get('code') != 0:
            logger.debug("No status data in bulb response")
            return
        self._interpret_apicall_result(inner_result)
        return

    def _interpret_apicall_result(self, response: dict) -> None:
        """Build detail dictionary from response."""
        self._brightness = response.get('brightness', 0)
        self._color_mode = response.get('colorMode', '')
        self._color = Color(red=response.get('red', 0),
                            green=response.get('green', 0),
                            blue=response.get('blue', 0))

    def set_brightness(self, brightness: int) -> bool:
        """Set brightness of bulb.

        Calls the `set_status` method with the brightness value.

        Args:
            brightness (int): Brightness of bulb (0-100).

        Returns:
            bool: True if successful, False otherwise.
        """
        return self.set_status(brightness=brightness)

    def set_rgb_color(self, red: NUMERIC_T, green: NUMERIC_T, blue: NUMERIC_T) -> bool:
        """DEPRECIATED, USE `set_rgb()`."""
        return self.set_status(red=red, green=green, blue=blue)

    def set_rgb(self, red: NUMERIC_T = None,
                green: NUMERIC_T = None,
                blue: NUMERIC_T = None) -> bool:
        return self.set_status(red=red, green=green, blue=blue)

    def set_hsv(self,
                hue: NUMERIC_T,
                saturation: NUMERIC_T,
                value: NUMERIC_T) -> bool | None:
        rgb = Color(hue=hue, saturation=saturation, value=value).rgb
        return self.set_status(red=rgb.red, green=rgb.green, blue=rgb.blue)

    def enable_white_mode(self) -> bool:
        """Enable white mode on bulb.

        Returns:
            bool: True if successful, False otherwise.
        """
        return self.set_status(brightness=100)

    def set_status(self, brightness: NUMERIC_T = None,
                   red: NUMERIC_T = None,
                   green: NUMERIC_T = None,
                   blue: NUMERIC_T = None) -> bool:
        """Set color of VeSync ESL100MC.

        Brightness or RGB values must be provided. If RGB values are provided,
        brightness is ignored.

        Args:
            brightness (int): Brightness of bulb (0-100).
            red (int): Red value of RGB color, 0-255.
            green (int): Green value of RGB color, 0-255.
            blue (int): Blue value of RGB color, 0-255.

        Returns:
            bool: True if successful, False otherwise.
        """
        brightness_update = 100
        if red is not None and green is not None and blue is not None:
            new_color = self._validate_rgb(red, green, blue)
            color_mode = 'color'
            if self.device_status == 'on' and new_color == self._color:
                logger.debug("New color is same as current color")
                return True
        else:
            logger.debug("RGB Values not provided")
            new_color = None
            if brightness is not None:
                brightness_update = int(self._validate_brightness(brightness))
                # Do nothing if brightness is passed and same as current
                if self.device_status == 'on' and brightness_update == self._brightness:
                    logger.debug('Brightness already set to %s', brightness)
                    return True
                color_mode = 'white'
            else:
                logger.debug("Brightness and RGB values are not set")
                return False

        head = helpers.bypass_header()
        body = helpers.bypass_body_v2(self.manager)
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'setLightStatus',
            'source': 'APP',
            'data': {
                'action': 'on',
                'speed': 0,
                'brightness': brightness_update,
                'red': 0 if new_color is None else int(new_color.rgb.red),
                'green': 0 if new_color is None else int(new_color.rgb.green),
                'blue': 0 if new_color is None else int(new_color.rgb.blue),
                'colorMode': 'color' if new_color is not None else 'white',
            }
        }

        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            method='post',
            headers=head,
            json_object=body,
        )
        if not isinstance(r, dict) or r.get('code') != 0:
            logger.debug("Error in setting bulb status")
            return False

        if color_mode == 'color' and new_color is not None:
            self._color_mode = 'color'
            self._color = Color(red=new_color.rgb.red,
                                green=new_color.rgb.green,
                                blue=new_color.rgb.blue)
        elif brightness is not None:
            self._brightness = int(brightness_update)
            self._color_mode = 'white'

        self.device_status = 'on'
        return True

    def toggle(self, status: str) -> bool:
        if status == 'on':
            turn_on = True
        elif status == 'off':
            turn_on = False
        else:
            logger.debug("Status must be on or off")
            return False
        head = helpers.bypass_header()
        body = helpers.bypass_body_v2(self.manager)
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'source': 'APP',
            'method': 'setSwitch',
            'data': {
                'id': 0,
                'enabled': turn_on
            }
        }
        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            method='post',
            headers=head,
            json_object=body,
        )
        if not isinstance(r, dict) or r.get('code') != 0:
            logger.debug("Error in setting bulb status")
            return False
        if turn_on is True:
            self.device_status = 'on'
        else:
            self.device_status = 'off'
        return True


class VeSyncBulbESL100(VeSyncBulb):
    """Object to hold VeSync ESL100 light bulb.

    This bulb only has the dimmable feature. Inherits from pyvesync.vesyncbulb.VeSyncBulb
    and pyvesync.vesyncbasedevice.VeSyncBaseDevice.

    Attributes:
        details (dict): Dictionary of bulb state details.
        brightness (int): Brightness of bulb (0-100).
        device_status (str): Status of bulb (on/off).
        connection_status (str): Connection status of bulb (online/offline).
    """

    def __init__(self, details: dict, manager: VeSync) -> None:
        """Initialize Etekcity ESL100 Dimmable Bulb.

        Args:
            details (dict): Dictionary of bulb state details.
            manager (VeSync): Manager class used to make API calls
        """
        super().__init__(details, manager)
        self.details: dict = {}

    def get_details(self) -> None:
        body = helpers.req_body(self.manager, 'devicedetail')
        body['uuid'] = self.uuid
        r, _ = helpers.call_api(
            '/SmartBulb/v1/device/devicedetail',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )
        if helpers.code_check(r):
            self.connection_status = r.get('connectionStatus')
            self.device_status = r.get('deviceStatus')
            if self.dimmable_feature:
                self._brightness = int(r.get('brightNess'))
        else:
            logger.debug('Error getting %s details', self.device_name)

    def get_config(self) -> None:
        body = helpers.req_body(self.manager, 'devicedetail')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid

        r, _ = helpers.call_api(
            '/SmartBulb/v1/device/configurations',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self.config = helpers.build_config_dict(r)
        else:
            logger.debug('Error getting %s config info', self.device_name)

    def toggle(self, status: str) -> bool:
        body = helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = status
        r, _ = helpers.call_api(
            '/SmartBulb/v1/device/devicestatus',
            'put',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )
        if helpers.code_check(r):
            self.device_status = status
            return True
        return False

    def set_brightness(self, brightness: int) -> bool:
        """Set brightness of dimmable bulb.

        Args:
            brightness (int): Brightness of bulb (0-100).

        Returns:
            bool: True if successful, False otherwise.
        """
        if not self.dimmable_feature:
            logger.debug('%s is not dimmable', self.device_name)
            return False
        brightness_update = int(self._validate_brightness(brightness))
        if self.device_status == 'on' and brightness_update == self._brightness:
            logger.debug("Device already in requested state")
            return True
        if self.device_status == 'off':
            self.toggle('on')
        body = helpers.req_body(self.manager, 'devicestatus')
        body['uuid'] = self.uuid
        body['status'] = 'on'
        body['brightNess'] = str(brightness_update)
        r, _ = helpers.call_api(
            '/SmartBulb/v1/device/updateBrightness',
            'put',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self._brightness = brightness_update
            return True

        logger.debug('Error setting brightness for %s', self.device_name)
        return False


class VeSyncBulbESL100CW(VeSyncBulb):
    """VeSync Tunable and Dimmable White Bulb."""

    def __init__(self, details: dict, manager: VeSync) -> None:
        """Initialize Etekcity Tunable white bulb."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        body = helpers.req_body(self.manager, 'bypass')
        body['cid'] = self.cid
        body['jsonCmd'] = {'getLightStatus': 'get'}
        body['configModule'] = self.config_module
        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/bypass',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )
        if not isinstance(r, dict) or not helpers.code_check(r):
            logger.debug('Error calling %s', self.device_name)
            return
        light_resp = r.get('result', {}).get('light')

        if light_resp is not None:
            self._interpret_apicall_result(light_resp)
        elif r.get('code') == -11300027:
            logger.debug('%s device offline', self.device_name)
            self.connection_status = 'offline'
            self.device_status = 'off'
        else:
            logger.debug(
                '%s - Unknown return code - %s with message %s',
                self.device_name,
                str(r.get('code', '')),
                str(r.get('msg', '')),
            )

    def _interpret_apicall_result(self, response: dict) -> None:
        self.connection_status = 'online'
        self.device_status = response.get('action', 'off')
        self._brightness = response.get('brightness', 0)
        self._color_temp = response.get('colorTempe', 0)

    def get_config(self) -> None:
        body = helpers.req_body(self.manager, 'bypass_config')
        body['uuid'] = self.uuid

        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/configurations',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self.config = helpers.build_config_dict(r)
        else:
            logger.debug('Error getting %s config info', self.device_name)

    def toggle(self, status: str) -> bool:
        if status not in ('on', 'off'):
            logger.debug('Invalid status %s', status)
            return False
        body = helpers.req_body(self.manager, 'bypass')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['jsonCmd'] = {'light': {'action': status}}
        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/bypass',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )
        if helpers.code_check(r):
            self.device_status = status
            return True
        logger.debug('%s offline', self.device_name)
        self.device_status = 'off'
        self.connection_status = 'offline'
        return False

    def set_brightness(self, brightness: int) -> bool:
        """Set brightness of tunable bulb."""
        brightness_update = int(self._validate_brightness(brightness))
        if self.device_status == 'on' and brightness_update == self._brightness:
            logger.debug("Device already in requested state")
            return True
        body = helpers.req_body(self.manager, 'bypass')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        light_dict: dict[str, NUMERIC_T] = {
            'brightness': brightness_update}
        if self.device_status == 'off':
            light_dict['action'] = 'on'
        body['jsonCmd'] = {'light': light_dict}
        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/bypass',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if helpers.code_check(r):
            self._brightness = brightness_update
            self.device_status = 'on'
            self.connection_status = 'online'
            return True
        self.device_status = 'off'
        self.connection_status = 'offline'
        logger.debug('%s offline', self.device_name)

        return False

    def set_color_temp(self, color_temp: int) -> bool:
        """Set Color Temperature of Bulb in pct (1 - 100)."""
        color_temp_update = self._validate_color_temp(color_temp)
        if self.device_status == 'on' and color_temp_update == self._color_temp:
            logger.debug("Device already in requested state")
            return True
        body = helpers.req_body(self.manager, 'bypass')
        body['cid'] = self.cid
        body['jsonCmd'] = {'light': {}}
        body['jsonCmd']['light']['colorTempe'] = color_temp_update
        if self.device_status == 'off':
            body['jsonCmd']['light']['action'] = 'on'
        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/bypass',
            'post',
            headers=helpers.req_headers(self.manager),
            json_object=body,
        )

        if not helpers.code_check(r):
            return False

        if r.get('code') == -11300027:
            logger.debug('%s device offline', self.device_name)
            self.connection_status = 'offline'
            self.device_status = 'off'
            return False
        if r.get('code') == 0:
            self.device_status = 'on'
            self._color_temp = color_temp
            return True
        logger.debug(
            '%s - Unknown return code - %d with message %s',
            self.device_name,
            r.get('code'),
            r.get('msg'),
        )
        return False


class VeSyncBulbValcenoA19MC(VeSyncBulb):
    """VeSync Multicolor Bulb."""

    def __init__(self, details: dict, manager: VeSync) -> None:
        """Initialize Multicolor bulb."""
        super().__init__(details, manager)

    def get_details(self) -> None:
        body = helpers.req_body(self.manager, 'bypassV2')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'getLightStatusV2',
            'source': 'APP',
            'data': {},
        }
        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=helpers.req_header_bypass(),
            json_object=body,
        )
        if not isinstance(r, dict) or not helpers.code_check(r):
            logger.debug('Error calling %s', self.device_name)
            return
        self._interpret_apicall_result(r)

    def _interpret_apicall_result(self, response: dict) -> None:
        if response.get('result', {}).get('result') is not None:
            innerresult = response.get('result', {}).get('result')
            self.connection_status = 'online'
            self.device_status = innerresult.get('enabled', 'off')
            if self.dimmable_feature:
                self._brightness = innerresult.get('brightness')
            if self.color_temp_feature:
                self._color_temp = innerresult.get('colorTemp')
            if self.rgb_shift_feature:
                self._color_mode = innerresult.get('colorMode')
                hue = float(round(innerresult.get('hue')/27.777777, 2))
                sat = float(innerresult.get('saturation')/100)
                val = float(innerresult.get('value'))
                self._color = Color(hue=hue, saturation=sat, value=val)
        elif (response.get('code') == -11300030 or
              response.get('code') == -11302030):
            logger.debug('%s device request timeout', self.device_name)
            self.connection_status = 'offline'
            self.device_status = 'off'
        elif response.get('code') == -11300027:
            logger.debug('%s device offline', self.device_name)
            self.connection_status = 'offline'
            self.device_status = 'off'
        else:
            logger.debug(
                '%s - Unknown return code - %d with message %s',
                self.device_name,
                response.get('code'),
                response.get('msg'),
            )

    def get_config(self) -> None:
        body = helpers.req_body(self.manager, 'bypass')
        body['method'] = 'configurations'
        body['uuid'] = self.uuid
        r, _ = helpers.call_api(
            '/cloud/v1/deviceManaged/configurations',
            'post',
            headers=helpers.req_header_bypass(),
            json_object=body,
        )
        if helpers.code_check(r):
            if r.get('result') is not None:
                result = r.get('result')
                self.__build_config_dict(result)
        else:
            logger.debug('Error getting %s config info', self.device_name)
            logger.debug('  return code - %d with message %s',
                         r.get('code'), r.get('msg'))

    def __build_config_dict(self, conf_dict: dict[str, str]) -> None:
        """Build configuration dict for Multicolor bulb."""
        self.config['currentFirmVersion'] = (
            conf_dict.get('currentFirmVersion', ''))
        self.config['latestFirmVersion'] = (
            conf_dict.get('latestFirmVersion', ''))
        self.config['firmwareUrl'] = (
            conf_dict.get('firmwareUrl', ''))
        self.config['allowNotify'] = (
            conf_dict.get('allowNotify', ''))
        self.config['deviceImg'] = (
            conf_dict.get('deviceImg', ''))
        self.config['defaultDeviceImg'] = (
            conf_dict.get('defaultDeviceImg', ''))
        self.config['ownerShip'] = (
            conf_dict.get('ownerShip', False))

    def toggle(self, status: str) -> bool:
        body = helpers.req_body(self.manager, 'bypassV2')
        if status == 'off':
            status_bool = False
        elif status == 'on':
            status_bool = True
        else:
            logger.debug('Invalid status %s for toggle - only on/off allowed',
                         status)
            return False
        if status == self.device_status:
            logger.debug("Device already in requested state")
            return True
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'setSwitch',
            'source': 'APP',
            'data': {
                'id': 0,
                'enabled': status_bool,
                }
            }
        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=helpers.req_header_bypass(),
            json_object=body)
        if helpers.code_check(r):
            self.device_status = status
            return True
        logger.debug('%s offline', self.device_name)
        self.device_status = 'off'
        self.connection_status = 'offline'
        return False

    def set_rgb(self, red: NUMERIC_T = None,
                green: NUMERIC_T = None,
                blue: NUMERIC_T = None) -> bool:
        new_color = Color(red=red, green=green, blue=blue).hsv
        return self.set_hsv(hue=new_color.hue,
                            saturation=new_color.saturation,
                            value=new_color.value)

    def set_brightness(self, brightness: int) -> bool:
        """Set brightness of multicolor bulb."""
        return self.set_status(brightness=brightness)

    def set_color_temp(self, color_temp: int) -> bool:
        """Set White Temperature of Bulb in pct (0 - 100)."""
        return self.set_status(color_temp=color_temp)

    def set_color_hue(self, color_hue: float) -> bool:
        """Set Color Hue of Bulb (0 - 360)."""
        return self.set_status(color_hue=color_hue)

    def set_color_saturation(self, color_saturation: float) -> bool:
        """Set Color Saturation of Bulb in pct (1 - 100)."""
        return self.set_status(color_saturation=color_saturation)

    def set_color_value(self, color_value: float) -> bool:
        """Set Value of multicolor bulb in pct (1 - 100)."""
        # Equivalent to brightness level, when in color mode.
        return self.set_status(color_value=color_value)

    def set_color_mode(self, color_mode: str) -> bool:
        """Set Color Mode of Bulb (white / hsv)."""
        return self.set_status(color_mode=color_mode)

    def set_hsv(self, hue: NUMERIC_T = None,
                saturation: NUMERIC_T = None,
                value: NUMERIC_T = None) -> bool:
        arg_dict = {
            "hue": self._validate_any(hue, 0, 360, 360) if hue is not None else "",
            "saturation": self._validate_any(
                saturation, 0, 100, 100) if saturation is not None else "",
            "brightness": self._validate_any(
                value, 0, 100, 100) if value is not None else ""
            }

        # the api expects the hsv Value in the brightness parameter
        if self._color is not None:
            current_dict = {"hue": self.color_hue,
                            "saturation": self.color_saturation,
                            "brightness": self.color_value}
            filtered_arg_dict = {k: v for k, v in arg_dict.items() if v != ""}
            same_colors = all(current_dict.get(k) == v
                              for k, v in filtered_arg_dict.items())
            if self.device_status == 'on' and same_colors:
                logger.debug("Device already in requested state")
                return True
        arg_dict = {
            "hue": int(round(arg_dict["hue"]*27.77778, 0)) if isinstance(
                arg_dict["hue"], float) else "",
            "saturation": int(round(arg_dict["saturation"]*100, 0)) if isinstance(
                arg_dict["saturation"], float) else "",
            "brightness": int(round(arg_dict["brightness"], 0)) if isinstance(
                arg_dict["brightness"], float) else ""
        }
        arg_dict['colorMode'] = 'hsv'
        return self._set_status_api(arg_dict)

    def enable_white_mode(self) -> bool:
        """Enable white color mode."""
        return self.set_status(color_mode='white')

    def set_status(self,  # noqa: C901
                   *,
                   brightness: NUMERIC_T = None,
                   color_temp: NUMERIC_T = None,
                   color_saturation: NUMERIC_T = None,
                   color_hue: NUMERIC_T = None,
                   color_mode: str | None = None,
                   color_value: NUMERIC_T = None
                   ) -> bool:
        """Set multicolor bulb parameters.

        No arguments turns bulb on. **Kwargs only**

        Args:
            brightness (int, optional): brightness between 0 and 100
            color_temp (int, optional): color temperature between 0 and 100
            color_mode (int, optional): color mode hsv or white
            color_hue (float, optional): color hue between 0 and 360
            color_saturation (float, optional): color saturation between 0 and 100
            color_value (int, optional): color value between 0 and 100

        Returns:
            bool : True if call was successful, False otherwise
        """
        arg_list = ['brightness', 'color_temp', 'color_saturation',
                    'color_hue', 'color_mode', 'color_value']
        toggle_on = True
        for val in arg_list:
            if locals()[val] is not None:
                toggle_on = False
        if toggle_on:
            self.turn_on()

        # If any HSV color values are passed,
        # set HSV status & ignore other values
        # Set Color if hue, saturation or value is set
        if any(var is not None for var in [color_hue, color_saturation, color_value]):
            return self.set_hsv(color_hue, color_saturation, color_value)

        # initiate variables
        request_dict = {
            "force": 1,
            "colorMode": '',
            "brightness": '',
            "colorTemp": '',
            "hue": "",
            "saturation": "",
            "value": ""
        }

        force_list = ['colorTemp', 'saturation', 'hue', 'colorMode', 'value']
        if brightness is not None:
            brightness_update = self._validate_brightness(brightness)
            if self.device_status == 'on' and brightness_update == self._brightness:
                logger.debug('Brightness already set to %s', brightness)
                return True
            if all(locals().get(k) is None for k in force_list):
                request_dict['force'] = 0
            request_dict['brightness'] = int(brightness_update)
        else:
            brightness_update = None
        # Set White Temperature of Bulb in pct (1 - 100).
        if color_temp is not None and \
                self._validate_any(color_temp, 0, 100, 100):
            valid_color_temp = self._validate_any(color_temp, 0, 100, 100)
            request_dict['colorTemp'] = int(valid_color_temp)
            request_dict['colorMode'] = 'white'

        # Set Color Mode of Bulb (white / hsv).
        if color_mode is not None:
            possible_modes = {'white': 'white',
                              'color': 'hsv',
                              'hsv': 'hsv'}
            if not isinstance(color_mode, str) or \
                    color_mode.lower() not in possible_modes:
                logger.error('Error: invalid color_mode value')
                return False
            color_mode = color_mode.lower()
            request_dict['colorMode'] = possible_modes[color_mode]
        if self._set_status_api(request_dict) and \
                brightness_update is not None:
            self._brightness = brightness_update
            return True
        return False

    def _set_status_api(self, data_dict: dict) -> bool:
        """Call api to set status - INTERNAL."""
        data_dict_start = {
            "force": 1,
            "brightness": '',
            "colorTemp": "",
            "colorMode": "",
            "hue": "",
            "saturation": "",
            "value": ""
        }
        data_dict_start.update(data_dict)
        body = helpers.req_body(self.manager, 'bypassV2')
        body['cid'] = self.cid
        body['configModule'] = self.config_module
        body['payload'] = {
            'method': 'setLightStatusV2',
            'source': 'APP',
            'data': data_dict_start
        }
        # Make API call
        r, _ = helpers.call_api(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=helpers.req_header_bypass(),
            json_object=body,
        )
        # Check result
        if helpers.code_check(r):
            self._interpret_apicall_result(r)
            self.device_status = 'on'
            return True
        self.device_status = 'off'
        self.connection_status = 'offline'
        logger.debug('%s offline', self.device_name)
        return False



"""
Light Bulbs Device API Responses

OUTLET variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'ESL100': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

# For a function to handle the response
def status_response(request_body=None):
    # do work with request_body
    return request_body, 200

METHOD_RESPONSES['ESL100']['set_status'] = status_response

# To change the default value for a device type

METHOD_RESPONSES['XYD0001'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)

"""
from copy import deepcopy
from utils import FunctionResponses, Defaults
from pyvesync.vesyncbulb import bulb_modules
TRACE_ID = "TRACE_ID"

# BULBS = ['ESL100', 'ESL100CW', 'ESL100MC', 'XYD0001']
BULBS = bulb_modules.keys()
BULBS_NUM = len(BULBS)


class BulbDetails:
    details_esl100 = ({
        'code': 0,
        'msg': None,
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'name': Defaults.name('ESL100'),
        'brightNess': Defaults.brightness,
        'timer': None,
        'away': None,
        'schedule': None,
        'ownerShip': '1',
        'scheduleCount': 0,
    }, 200)

    details_esl100cw = ({
        "traceId": TRACE_ID,
        "code": 0,
        "msg": None,
        "module": None,
        "stacktrace": None,
        "result": {
            "light": {
                "action": "on",
                "brightness": Defaults.brightness,
                "colorTempe": Defaults.color_temp,
            }
        }
    }, 200)

    details_esl100mc = ({
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "action": "on",
                "brightness": Defaults.brightness,
                "colorMode": "color",
                "speed": 0,
                "red": Defaults.color.rgb.red,
                "green": Defaults.color.rgb.green,
                "blue": Defaults.color.rgb.blue,
            }
        }
    }, 200)

    details_valceno = (
            {
                "traceId": TRACE_ID,
                "code": 0,
                "msg": "request success",
                "result": {
                    "traceId": TRACE_ID,
                    "code": 0,
                    "result": {
                        "enabled": "on",
                        "colorMode": "color",
                        "brightness": Defaults.brightness,
                        "colorTemp": Defaults.color_temp,
                        "hue": Defaults.color.hsv.hue*27.7778,
                        "saturation": Defaults.color.hsv.saturation*100,
                        "value": Defaults.color.hsv.value,
                    }
                }
            }, 200
        )


DETAILS_RESPONSES = {
    'ESL100': BulbDetails.details_esl100,
    'ESL100CW': BulbDetails.details_esl100cw,
    'ESL100MC': BulbDetails.details_esl100mc,
    'XYD0001': BulbDetails.details_valceno,
}


def valceno_set_status_response(kwargs=None):
    default_resp = {
        "traceId": Defaults.trace_id,
        "code": 0,
        "msg": "request success",
        "result": {
            "traceId": Defaults.trace_id,
            "code": 0,
            "result": {
                "enabled": "on",
                "colorMode": "hsv",
                "brightness": Defaults.brightness,
                "colorTemp": Defaults.color_temp,
                "hue": Defaults.color.hsv.hue*27.7778,
                "saturation": Defaults.color.hsv.saturation*100,
                "value": Defaults.color.hsv.value
            }
        }
    }
    if kwargs is not None and isinstance(kwargs, dict):
        if kwargs.get('hue') is not None:
            default_resp['result']['result']['hue'] = kwargs['hue'] * 27.7778
        if kwargs.get('saturation') is not None:
            default_resp['result']['result']['saturation'] = kwargs['saturation'] * 100
        if kwargs.get('value') is not None:
            default_resp['result']['result']['value'] = kwargs['value']
    return default_resp, 200


METHOD_RESPONSES = {k: deepcopy(FunctionResponses) for k in BULBS}

METHOD_RESPONSES['XYD0001'].default_factory = lambda: ({
                                                "traceId": Defaults.trace_id,
                                                "code": 0,
                                                "msg": "request success",
                                                "result": {
                                                    "traceId": Defaults.trace_id,
                                                    "code": 0
                                                }
                                            }, 200)

METHOD_RESPONSES['ESL100MC'].default_factory = lambda: ({
                                                "traceId": Defaults.trace_id,
                                                "code": 0,
                                                "msg": "request success",
                                                "result": {
                                                    "traceId": Defaults.trace_id,
                                                    "code": 0
                                                }
                                            }, 200)

XYD0001_RESP = {
    'set_brightness': valceno_set_status_response,
    'set_color_temp': valceno_set_status_response,
    'set_hsv': valceno_set_status_response,
    'set_rgb': valceno_set_status_response,
}

METHOD_RESPONSES['XYD0001'].update(XYD0001_RESP)



"""
Outlet Device API Responses

OUTLET variable is a list of device types

DETAILS_RESPONSES variable is a dictionary of responses from the API
for get_details() methods.  The keys are the device types and the
values are the responses.  The responses are tuples of (response, status)

METHOD_RESPONSES variable is a defaultdict of responses from the API. This is
the FunctionResponse variable from the utils module in the tests dir.
The default response is a tuple with the value ({"code": 0, "msg": "success"}, 200).

The values of METHOD_RESPONSES can be a function that takes a single argument or
a static value. The value is checked if callable at runtime and if so, it is called
with the provided argument. If not callable, the value is returned as is.

METHOD_RESPONSES = {
    'DEV_TYPE': defaultdict(
        lambda: ({"code": 0, "msg": "success"}, 200))
    )
}

### For a function to handle the response
def status_response(request_kwargs=None):
    # do work with request_kwargs
    return request_body, 200

METHOD_RESPONSES['DEV_TYPE']['set_status'] = status_response

### To change the default value for a device type

METHOD_RESPONSES['DEVTYPE'].default_factory = lambda: ({"code": 0, "msg": "success"}, 200)

If changing default response for all devices, change the default factory of the import
default dict but make sure to use `deepcopy` to avoid unintended side effects.
"""
from copy import deepcopy
from collections import defaultdict
from utils import FunctionResponses, Defaults
from pyvesync.vesyncoutlet import outlet_modules

OUTLETS = outlet_modules.keys()
OUTLETS_NUM = len(OUTLETS)

# OUTLETS = ['wifi-switch-1.3', 'ESW03-USA', 'ESW01-EU', 'ESW15-USA', 'ESO15-TB']


class OutletDefaults:
    voltage = 1  # volts
    energy = 1  # kilowatt
    power = 1  # kilowatt-hours
    round_7a_voltage = '1000:1000'  # 1 volt
    round_7a_power = '1000:1000'  # 1 watt


class OutletDetails:
    details_15a = (
        {
            'code': 0,
            'msg': None,
            'deviceStatus': 'on',
            'connectionStatus': 'online',
            'activeTime': Defaults.active_time,
            'energy': 1,
            'nightLightStatus': 'on',
            'nightLightBrightness': 50,
            'nightLightAutomode': 'manual',
            'power': '1',
            'voltage': '1',
        },
        200,
    )

    details_7a = (
        {
            'deviceStatus': 'on',
            'deviceImg': '',
            'activeTime': Defaults.active_time,
            'energy': OutletDefaults.energy,
            'power': OutletDefaults.round_7a_power,
            'voltage': OutletDefaults.round_7a_voltage,
        },
        200,
    )

    details_10a = (
        {
            'code': 0,
            'msg': None,
            'deviceStatus': 'on',
            'connectionStatus': 'online',
            'activeTime': Defaults.active_time,
            'energy': OutletDefaults.energy,
            'nightLightStatus': None,
            'nightLightBrightness': None,
            'nightLightAutomode': None,
            'power': OutletDefaults.power,
            'voltage': OutletDefaults.voltage,
        },
        200,
    )

    details_outdoor = (
        {
            'code': 0,
            'msg': None,
            'connectionStatus': 'online',
            'activeTime': Defaults.active_time,
            'energy': OutletDefaults.energy,
            'power': OutletDefaults.power,
            'voltage': OutletDefaults.voltage,
            'deviceStatus': 'on',
            'deviceName': Defaults.name('ESO15-TB'),
            'subDevices': [
                {
                    'subDeviceNo': 1,
                    'defaultName': 'Socket A',
                    'subDeviceName': Defaults.name('ESO15-TB'),
                    'subDeviceStatus': 'on',
                },
                {
                    'subDeviceNo': 2,
                    'defaultName': 'Socket B',
                    'subDeviceName': Defaults.name('ESO15-TB'),
                    'subDeviceStatus': 'on',
                },
            ],
        },
        200,
    )

    bsdgo1_details = ({
        "code": 0,
        "msg": "request success",
        "result": {
            "powerSwitch_1": 1,
            "traceId": "1735308365651",
            "active_time": Defaults.active_time,
            "connectionStatus": "online",
            "code": 0
        }
    }, 200)


DETAILS_RESPONSES = {
    'wifi-switch-1.3': OutletDetails.details_7a,
    'ESW03-USA': OutletDetails.details_10a,
    'ESW01-EU': OutletDetails.details_10a,
    'ESW15-USA': OutletDetails.details_15a,
    'ESO15-TB': OutletDetails.details_outdoor,
    'BSDOG01': OutletDetails.bsdgo1_details
}

ENERGY_HISTORY = (
    {
        'code': 0,
        'energyConsumptionOfToday': 1,
        'costPerKWH': 1,
        'maxEnergy': 1,
        'totalEnergy': 1,
        'data': [
            1,
            1,
        ],
    },
    200,
)


METHOD_RESPONSES = {k: deepcopy(FunctionResponses) for k in OUTLETS}

for k in METHOD_RESPONSES:
    METHOD_RESPONSES[k]['get_weekly_energy'] = ENERGY_HISTORY
    METHOD_RESPONSES[k]['get_monthly_energy'] = ENERGY_HISTORY
    METHOD_RESPONSES[k]['get_yearly_energy'] = ENERGY_HISTORY

# Add BSDGO1 specific responses
METHOD_RESPONSES['BSDOG01'] = defaultdict(lambda: ({
    "code": 0,
    "msg": "request success",
    "result": {
        "traceId": Defaults.trace_id,
        "code": 0
    }
}, 200))



"""pyvesync setup script."""

from os import path
from setuptools import setup, find_packages


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pyvesync',
    version='2.1.18',
    description='pyvesync is a library to manage Etekcity\
                 Devices, Cosori Air Fryers and Levoit Air \
                     Purifiers run on the VeSync app.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/webdjoe/pyvesync',
    author='Mark Perdue, Joe Trabulsy',
    author_email='webdjoe@gmail.com',
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Home Automation',
    ],
    keywords=['iot', 'vesync', 'levoit', 'etekcity', 'cosori', 'valceno'],
    packages=find_packages('src', exclude=['tests', 'test*']),
    package_dir={'': 'src'},
    install_requires=['requests>=2.20.0'],
    extras_require={
        'dev': ['pytest', 'pytest-cov', 'pyyaml', 'tox']
    },
    python_requires='>=3.9',
)



"""Base class for all VeSync devices."""
from __future__ import annotations
import logging
import json
from typing import TYPE_CHECKING
from pyvesync.helpers import Helpers as helper  # noqa: N813
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from pyvesync import VeSync


class VeSyncBaseDevice:
    """Properties shared across all VeSync devices.

    Base class for all VeSync devices.

    Parameters:
        details (dict): Device details from API call.
        manager (VeSync): Manager object for API calls.

    Attributes:
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_status (str): Connection status of device.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        mode (str): Mode of device.
        speed (Union[str, int]): Speed of device.
        extension (dict): Extension of device, not used.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        config (dict): Configuration of device, including firmware version
        device_status (str): Status of device, on or off.

    Methods:
        is_on(): Return True if device is on.
        firmware_update(): Return True if firmware update available.
        display(): Print formatted device info to stdout.
        displayJSON(): JSON API for device details.
    """

    def __init__(self, details: dict, manager: VeSync) -> None:
        """Initialize VeSync device base class."""
        self.manager = manager
        if 'cid' in details and details['cid'] is not None:
            self.device_name: str = details['deviceName']
            self.device_image: str | None = details.get('deviceImg')
            self.cid: str = details['cid']
            self.connection_status: str = details['connectionStatus']
            self.connection_type: str | None = details.get(
                'connectionType')
            self.device_type: str = details['deviceType']
            self.type: str | None = details.get('type')
            self.uuid: str | None = details.get('uuid')
            self.config_module: str = details['configModule']
            self.mac_id: str | None = details.get('macID')
            self.mode: str | None = details.get('mode')
            self.speed: int | None = details.get('speed') if details.get(
                'speed') != '' else None
            self.extension = details.get('extension')
            self.current_firm_version = details.get(
                    'currentFirmVersion')
            self.device_region: str | None = details.get('deviceRegion')
            self.pid = None
            self.sub_device_no = details.get('subDeviceNo', 0)
            self.config: dict = {}
            if isinstance(details.get('extension'), dict):
                ext = details['extension']
                self.speed = ext.get('fanSpeedLevel')
                self.mode = ext.get('mode')
            if self.connection_status != 'online':
                self.device_status: str | None = 'off'
            else:
                self.device_status = details.get('deviceStatus')

        else:
            logger.error('No cid found for %s', self.__class__.__name__)

    def __eq__(self, other: object) -> bool:
        """Use device CID and subdevice number to test equality."""
        if not isinstance(other, VeSyncBaseDevice):
            return NotImplemented
        return bool(other.cid == self.cid
                    and other.sub_device_no == self.sub_device_no)

    def __hash__(self) -> int:
        """Use CID and sub-device number to make device hash."""
        if isinstance(self.sub_device_no, int) and self.sub_device_no > 0:
            return hash(self.cid + str(self.sub_device_no))
        return hash(self.cid)

    def __str__(self) -> str:
        """Use device info for string represtation of class."""
        return f'Device Name: {self.device_name}, \
                 Device Type: {self.device_type},\
                 SubDevice No.: {self.sub_device_no},\
                 Status: {self.device_status}'

    def __repr__(self) -> str:
        """Representation of device details."""
        return f'DevClass: {self.__class__.__name__},\
                Name:{self.device_name}, Device No: {self.sub_device_no},\
                DevStatus: {self.device_status}, CID: {self.cid}'

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        return (self.device_status == 'on')

    @property
    def firmware_update(self) -> bool:
        """Return True if firmware update available."""
        cfv = self.config.get('current_firmware_version')
        lfv = self.config.get('latest_firmware_version')
        if cfv is not None and lfv is not None:
            if cfv != lfv:
                return True
        else:
            logger.debug('Call device.get_config() to get firmware versions')
        return False

    def get_pid(self) -> None:
        """Get managed device configuration."""
        body = helper.req_body(self.manager, 'devicedetail')
        body['configModule'] = self.config_module
        body['region'] = self.device_region
        body['method'] = 'configInfo'
        r, _ = helper.call_api('/cloud/v1/deviceManaged/configInfo',
                               'post',
                               json_object=body)
        if not isinstance(r, dict) or r.get('code') != 0 or r.get('result') is None:
            logger.error('Error getting config info for %s', self.device_name)
            return
        self.pid = r.get('result', {}).get('pid')

    def display(self) -> None:
        """Print formatted device info to stdout.

        Example:
        ```
        Device Name:..................Living Room Lamp
        Model:........................ESL100
        Subdevice No:.................0
        Status:.......................on
        Online:.......................online
        Type:.........................wifi
        CID:..........................1234567890abcdef
        ```
        """
        disp = [
            ('Device Name:', self.device_name),
            ('Model: ', self.device_type),
            ('Subdevice No: ', str(self.sub_device_no)),
            ('Status: ', self.device_status),
            ('Online: ', self.connection_status),
            ('Type: ', self.type),
            ('CID: ', self.cid),
        ]
        if self.uuid is not None:
            disp.append(('UUID: ', self.uuid))

        for line in disp:
            print(f'{line[0]:.<30} {line[1]}')

    def displayJSON(self) -> str:  # pylint: disable=invalid-name
        """JSON API for device details.

        Returns:
            str: JSON formatted string of device details.

        Example:
        ```
        {
            "Device Name": "Living Room Lamp",
            "Model": "ESL100",
            "Subdevice No": "0",
            "Status": "on",
            "Online": "online",
            "Type": "wifi",
            "CID": "1234567890abcdef"
        }
        ```
        """
        return json.dumps(
            {
                'Device Name': self.device_name,
                'Model': self.device_type,
                'Subdevice No': str(self.sub_device_no),
                'Status': self.device_status,
                'Online': self.connection_status,
                'Type': self.type,
                'CID': self.cid,
            },
            indent=4)



def pytest_addoption(parser):
    """Prevent new API's from being written during pipeline tests."""
    parser.addoption(
        "--write_api", action="store_true", default=False,
          help="run tests without writing API to yaml"
    )
    parser.addoption(
        "--overwrite", action="store_true", default=False,
          help="overwrite existing API in yaml - WARNING do not use unless absolutely necessary"
    )

def pytest_generate_tests(metafunc):
    if metafunc.cls is None or 'test_x' in metafunc.module.__name__:
        return
    if metafunc.config.getoption('--write_api'):
        write_api = True
    else:
        write_api = False
    if metafunc.config.getoption('--overwrite'):
        overwrite = True
    else:
        overwrite = False
    metafunc.cls.overwrite = overwrite
    metafunc.cls.write_api = write_api
    if 'device' in metafunc.cls.__dict__:
        device = metafunc.cls.__dict__['device']
        if device not in metafunc.cls.__dict__:
            return
        devices = metafunc.cls.__dict__[device]
        if metafunc.function.__name__ == 'test_details':
            return details_generator(metafunc, device, devices)

        elif metafunc.function.__name__ == 'test_methods':
            return method_generator(metafunc, device, devices)


def details_generator(metafunc, gen_type, devices):
    """Parametrize device tests for get_details()."""
    id_list = []
    argvalues = []
    for dev_type in devices:
        id_list.append(f"{gen_type}.{dev_type}.update")
        argvalues.append([dev_type, 'update'])
    metafunc.parametrize("dev_type, method", argvalues, ids=id_list)
    return


def method_generator(metafunc, gen_type, devices):
    """Parametrize device tests for methods."""
    if 'base_methods' in metafunc.cls.__dict__:
        base_methods = metafunc.cls.__dict__['base_methods']
    else:
        base_methods = []
    call_dict = {dt: base_methods.copy() for dt in devices}
    if 'device_methods' in metafunc.cls.__dict__:
        dev_methods = metafunc.cls.__dict__['device_methods']
        for dev in call_dict:
            if dev in dev_methods:
                call_dict[dev].extend(dev_methods[dev])
    id_list = []
    argvalues = []
    for dev, methods in call_dict.items():
        for method in methods:
            id_list.append(f'{gen_type}.{dev}.{method}')
            argvalues.append([dev, method])
    metafunc.parametrize("dev_type, method", argvalues, ids=id_list)
    return


"""Test scripts for BSDGO1 Outlets."""

import pytest
from unittest.mock import patch
import logging
from pyvesync import VeSync
from pyvesync.vesyncoutlet import VeSyncOutletBSDGO1
from pyvesync.helpers import Helpers as helpers
import call_json
import call_json_outlets
from utils import Defaults, TestBase

DEVICE_TYPE = 'BSDOG01'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_TYPE)

CORRECT_BSDGO1_LIST = call_json.DeviceList.device_list_response(DEVICE_TYPE)

CORRECT_BSDGO1_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEVICE_TYPE]

BAD_BSDGO1_LIST = call_json.DETAILS_BADCODE


class TestVeSyncBSDGO1Switch(TestBase):
    """Test BSDGO1 outlet API."""

    def test_bsdgo1_conf(self):
        """Test initialization of BSDGO1 outlet."""
        self.mock_api.return_value = CORRECT_BSDGO1_LIST
        self.manager.get_devices()
        outlets = self.manager.outlets
        assert len(outlets) == 1
        bsdgo1_outlet = outlets[0]
        assert isinstance(bsdgo1_outlet, VeSyncOutletBSDGO1)
        assert bsdgo1_outlet.device_name == call_json.Defaults.name(DEVICE_TYPE)
        assert bsdgo1_outlet.device_type == DEVICE_TYPE
        assert bsdgo1_outlet.cid == call_json.Defaults.cid(DEVICE_TYPE)
        assert bsdgo1_outlet.uuid == call_json.Defaults.uuid(DEVICE_TYPE)

    def test_bsdgo1_details(self):
        """Test BSDGO1 get_details()."""
        self.mock_api.return_value = CORRECT_BSDGO1_DETAILS
        bsdgo1_outlet = VeSyncOutletBSDGO1(DEV_LIST_DETAIL, self.manager)
        bsdgo1_outlet.get_details()
        response = CORRECT_BSDGO1_DETAILS[0]
        result = response.get('result', {})
        
        expected_status = 'on' if result.get('powerSwitch_1') == 1 else 'off'
        assert bsdgo1_outlet.device_status == expected_status
        
        assert result.get('active_time') == Defaults.active_time
        assert result.get('connectionStatus') == 'online'

    def test_bsdgo1_details_fail(self):
        """Test BSDGO1 get_details with bad response."""
        self.mock_api.return_value = BAD_BSDGO1_LIST
        bsdgo1_outlet = VeSyncOutletBSDGO1(DEV_LIST_DETAIL, self.manager)
        bsdgo1_outlet.get_details()
        assert len(self.caplog.records) == 1
        assert 'details' in self.caplog.text

    def test_bsdgo1_onoff(self):
        """Test BSDGO1 Device On/Off Methods."""
        self.mock_api.return_value = ({'code': 0}, 200)
        bsdgo1_outlet = VeSyncOutletBSDGO1(DEV_LIST_DETAIL, self.manager)
        head = helpers.req_header_bypass()
        body = helpers.req_body(self.manager, 'bypassV2')
        body['cid'] = bsdgo1_outlet.cid
        body['configModule'] = bsdgo1_outlet.config_module
        
        # Test turn_on
        body['payload'] = {
            'data': {'powerSwitch_1': 1},
            'method': 'setProperty',
            'source': 'APP'
        }
        on = bsdgo1_outlet.turn_on()
        self.mock_api.assert_called_with(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=head,
            json_object=body
        )
        assert on
        
        # Test turn_off
        body['payload'] = {
            'data': {'powerSwitch_1': 0},
            'method': 'setProperty',
            'source': 'APP'
        }
        off = bsdgo1_outlet.turn_off()
        self.mock_api.assert_called_with(
            '/cloud/v2/deviceManaged/bypassV2',
            'post',
            headers=head,
            json_object=body
        )
        assert off

    def test_bsdgo1_onoff_fail(self):
        """Test BSDGO1 On/Off Fail with bad response."""
        self.mock_api.return_value = BAD_BSDGO1_LIST
        bsdgo1_outlet = VeSyncOutletBSDGO1(DEV_LIST_DETAIL, self.manager)
        assert not bsdgo1_outlet.turn_on()
        assert not bsdgo1_outlet.turn_off() 